/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

package MetaTF.Generator.Java;

import MetaTF.Parser.*;
import MetaTF.Types;
import MetaTF.Panic;
import MetaTF.Generator.Java.Util;
import MetaTF.Generator.Java.ConstructorHeaderGenerator;
import MetaTF.Generator.Java.ConstructorBodyGenerator;
import MetaTF.Generator.Java.PutRecordGenerator;
import java.util.*;
import java.io.*;

/***********************************************************************************************
 * ClassGenerator.
 * Generate Java classes for each user-defined section record.
 * These classes all extend Section or Record.
 *
 * @see <a href="../../Parser/metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.3 $, $Date: 2001/03/29 15:53:12 $),
 * @author Richard Jones
**/
class ClassGenerator implements DTDVisitor {

    /** The name of the dtd **/
    private final String dtd;

   /** The parent directory **/
    private final File parentDir;
  
    /** The name of the main class **/
    private final String className;

    /** The current section **/
    private int sectionNumber;

    /** The symbol table **/
    private DTD symtab;


    /** The FileWriter to which output is to be written **/
    private FileWriter writer;

    /**
     * Allocate a new object to write putRecord methods
     *
     * @param dtd the name of the DTD 
     * @param cname the name of the class being generated     
     * @param w the FileWrite to which the method definition is to be written
    **/
    ClassGenerator(String dtd, File parent, String cname) {
        this.dtd = dtd;
        parentDir = parent;
        className = cname;
    }
 
    
    /**
     * Visit the symbol table to generate classes for each user-defined record
     * @param symtab the symbol table
     * @param data not used
     * @return null
    **/
    public Object visit(DTD symtab, Object data) {
        this.symtab = symtab;
        /* Emit code in ascending numerical order of sections */
        Map sections = symtab.getSections();
        Set keys = new TreeSet(sections.keySet()); 
        Iterator it = keys.iterator();
        while(it.hasNext()) {
            Integer sno = (Integer)it.next();
            sectionNumber = sno.intValue();
            ((Section)sections.get(sno)).accept(this, data);
        }
        return null;
    }

    String sectionName = null;
    
    public Object visit(Section section, Object data) {
        sectionName = section.getName();
        boolean cached = section.isCached();
        sectionNumber = section.getNumber();

        /*
         * 1. Set up the writer for <section name>.java
         */
        try {
            writer = new FileWriter(new File(parentDir, Util.classify(sectionName)+".java"));
        } catch(IOException e) {
            Panic.panic(e.getMessage());
        }

        /*
         * 2. Emit the section class, including record cache (if any)
         */
        out(
"package "+className+";                                                                           \n\n"+
"import MetaTF.BadRecordException;                                                                \n"+
"import MetaTF.Runtime.*;                                                                         \n"+
"import java.io.IOException;                                                                      \n"+
"import java.util.*;                                                                              \n");

        if(cached)
           out(
"/** This class represents "+dtd+" cached section "+sectionName+" "+sectionNumber+" **/             ");

        else
           out(
"/** This class represents "+dtd+" section "+sectionName+" "+sectionNumber+" **/                    ");

        out(
"public class "+Util.classify(sectionName)+" extends MetaTF.Runtime.Section {                 \n\n"+
"    public static final int sectionNumber = "+sectionNumber+";                               \n\n"+
"    public "+Util.classify(sectionName)+"() {                                                \n"+
"    }                                                                                        \n");

        if(cached) {
            out(

"    /** This section is cached **/                                                           \n"+            
"    public final boolean iscached = true;                                                    \n\n"+

"    /* The record cache */                                                                   \n"+
"    private static List records = new ArrayList();                                           \n\n"+

"    /** Add a record to the cache **/                                                        \n"+
"    public static void cache(Record r) { records.add(r); }                                          \n\n"+

"    /**                                                                                      \n"+
"     * Retrieve a record from the cache                                                      \n"+
"     * @param i the number of the record to be retrieved                                     \n"+
"     * @exception BadRecordException IF i<0 || i > size of the cache                         \n"+
"    **/                                                                                      \n"+
"    public static Record getRecord(int i) throws BadRecordException {                               \n"+
"        if(i<0)                                                                              \n"+
"            return null;                                                                     \n"+
"        if(i > records.size())                                                               \n"+
"            throw new BadRecordException(\""+Util.classify(sectionName)+".getRecord got bad offset \"+i);  \n"+
"        return (Record)records.get(i);                                                       \n"+
"    }                                                                                        \n\n"+

"    /** Retrive the whole cache **/                                                          \n"+
"    public static List getCache() { return records; }                                               \n");

        } else
            out(
"    public final static boolean iscached = false;                                            \n");

        out(
"    public static void putMetadata(LowLevelWriter llw, int f, int i) throws BadRecordException, IOException {  \n"+
"        Section.putSectionMetadata(llw, sectionNumber, f, i);                                       \n"+           
"    }                                                                                        \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, long v) throws BadRecordException, IOException { \n"+
"        Section.putSectionMetadata(llw, sectionNumber, f, i, v);                                    \n"+           
"    }                                                                                        \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, long v1, long v2) throws BadRecordException, IOException { \n"+
"        Section.putSectionMetadata(llw, sectionNumber, f, i, v1, v2);                               \n"+           
"    }                                                                                        \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, String v) throws BadRecordException, IOException { \n"+
"        Section.putSectionMetadata(llw, sectionNumber, f, i, v);                                    \n"+           
"    }                                                                                        \n\n"+

"} // end of class "+Util.classify(sectionName)+" (section "+sectionName+" "+sectionNumber+")                   ");

        /* Finished generating <section name>.java */
        try {
            writer.flush();
            writer.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
        }


        /*
         * 3. Emit class for each user-defined record, tagged or not
         */
        Map records = section.getRecords();
        Set keys = new TreeSet(records.keySet());
        Iterator it = keys.iterator();
        while(it.hasNext()) {
            String s = (String)it.next();
            ((Record)records.get(s)).accept(this,data);
        }

        return null;
    }


    /**
     *  The names of fields encountered in this record (and its component types)
    **/
    private Map knownFields;

    /** The number of fields encountered in this record (and its component types) **/
    private int fieldCount;
    
    public Object visit(Record record, Object data) {
        String name = record.getName();
        String fullname = Util.classify(sectionName)+"_"+Util.classify(name);

        /*
         * 1. Set up the writer for <record name>.java
         */
        try {
            writer = new FileWriter(new File(parentDir,fullname+".java"));
        } catch(IOException e) {
            Panic.panic(e.getMessage());
        }

        /*
         * 2. Emit class for this user-defined record
         */
        out(
"package "+className+";                                                                         \n\n"+
"import MetaTF.BadRecordException;                                                              \n"+
"import MetaTF.Runtime.*;                                                                      \n"+
"import java.io.*;                                                                              \n\n"+

"/** This class represents "+name+" records in section "+sectionName+" "+sectionNumber+" of "+dtd+" **/ \n"+
"public class "+fullname+" extends MetaTF.Runtime.Record {                    \n"+

"    public static final int sectionNumber = "+sectionNumber+";                         \n");

        Field f = record.getField();
        knownFields = new HashMap(); 
        fieldCount = 0;
        Iterator it;
        if((f instanceof NamedField) || (f instanceof RepeatField)) {
            out(            
"    "+(String)f.accept(this, knownFields)+";"                                             );

        } else {
            /*
             * f instanceof TupleField - must treat outer-level tuples specially
             */
            it = ((TupleField)f).getFields().iterator();
            while(it.hasNext()) {
                Field f2 = (Field)it.next();
                String s = (String)f2.accept(this, knownFields);
                if(s!=null)
                    out(
"    "+s+";                                                                              ");
            }
        }

        /* 
         * 3. Generate constructors for this record
         *    Need (a) constructor with no arguments and (b) constructor with initializers for all members
         *    First, reset knownfields before generating the non-trivial constructor
         */

        // The trivial constructor
        out(
"\n    /** Allocate a new object representing record "+name+" in section "+sectionNumber+" **/ \n"+        
"    public "+fullname+"() {}                                                 \n");

        // The non-trivial constructor
        ConstructorHeaderGenerator chg = new ConstructorHeaderGenerator(symtab, sectionNumber); // the header        
        ConstructorBodyGenerator   cbg = new ConstructorBodyGenerator(symtab, sectionNumber);   // the body
        knownFields = new HashMap(); 
        String header;
        String body = null;
        
        if(f instanceof NamedField) {
            header = (String)f.accept(chg, knownFields);
            if(header!=null) {
                knownFields = new HashMap(); 
                body = (String)f.accept(cbg, knownFields)+
"    }                                                                                   ";
            }
        }
        
        else if(f instanceof TagField) {
            header = null;
            body = null;
        }
        
        else if(f instanceof RepeatField) {
            header = (String)f.accept(chg, knownFields);
            if(header!=null) {
                knownFields = new HashMap(); 
                body = (String)f.accept(cbg, knownFields)+
"    }                                                                              ";
            }
        }
        
        else { // f instanceof TupleField
            it = ((TupleField)f).getFields().iterator();
            header = null; // shut the compiler up
            String tmp = null;
            while(tmp==null && it.hasNext()) {
                Field f2 = (Field)it.next();
                tmp = (String)f2.accept(chg, knownFields);	
            }
            if(tmp!=null) {
                header = tmp;
                while(it.hasNext()) {
                    Field f2 = (Field)it.next();
                    tmp = (String)f2.accept(chg, knownFields);
                    if(tmp!=null)
                        header += ", "+tmp;	
                }
            }
            if(header!=null) {
                knownFields = new HashMap(); 
                body = null;
                it = ((TupleField)f).getFields().iterator();
                while(body==null && it.hasNext()) {
                    Field f2 = (Field)it.next();
                    body = (String)f2.accept(cbg, knownFields);	
                }
                if(body!=null) {
                    while(it.hasNext()) {
                        Field f2 = (Field)it.next();
                        tmp = (String)f2.accept(cbg, knownFields);
                        if(tmp!=null)
                            body += "\n"+tmp;
                    }
                }
                body += "\n    }";
            }
        }
        if(header!=null) {
            out(
"    /** Allocate a new object representing record "+name+" in section "+sectionName+" "+sectionNumber+" **/  \n"+        
"    public "+fullname+"("+header+ ") {                                            ");
            out(body);
        }

        
        /* 
         * 4. Generate putRecord method for this record
         *    First, reset knownfields before generating the putRecords to make sure field counting is correct
         */
        knownFields = new HashMap(); 
        int tag;
        try {
            tag = record.getTag();
            out("\n"+
"    public void putRecord(LowLevelWriter llw) throws IOException {                          \n"+
"        TraceRecord tr = MetaTFState.getTraceRecord("+sectionNumber+", "+tag+");            \n"+
"        TraceField tf;                                                                      ");

            PutRecordGenerator prg = new PutRecordGenerator(writer);
            if(f instanceof TupleField ) {
                it = ((TupleField)f).getFields().iterator();
                while(it.hasNext()) {
                    Field f2 = (Field)it.next();
                    String s = (String)f2.accept(prg, knownFields);	// generate putRecord()
                }
            } else {
                f.accept(prg, knownFields);
            } 
            out(
"        llw.writeln();                                                                     \n"+
"    }                                                                                      \n");

            /*
             * 5. Emit putMetadata methods for this record
             */
            out(
"    /**                                                                                    \n"+
"     * Put a record to change the encoding of a field to None                        \n"+
"     *                                                                                     \n"+
"     * @param llw the writer                                                               \n"+
"     * @param f the field to be changed; this should be MetaTFConstants.None               \n"+
"     * @param i the encoding                                                         \n"+
"     * @exception BadRecord IF f!=MetaTFConstants.None                                     \n"+
"     * @exception IOException IF the write fails                                           \n"+
"     * @see MetaTF.MetaTFConstants                                                         \n"+
"    **/                                                                                    \n"+
"    static void putMetadata(LowLevelWriter llw, int f, int i) throws BadRecordException, IOException {  \n"+
"        Record.putMetadata(llw, sectionNumber, tag, f, i);                                 \n"+
"    }                                                                                      \n\n"+
    
"    /**                                                                                    \n"+
"     * Put a record to change the encoding of a field to a single, long-valued encoding  \n"+
"     *                                                                                     \n"+
"     * @param llw the writer                                                               \n"+
"     * @param f the field to be changed                                                    \n"+
"     * @param i the encoding; this should be one of MetaTFConstants.BaseOffset,      \n"+
"     *          MetaTFConstants.Delta, MetaTFConstants.Default, MetaTFConstants.SectionOffset  \n"+
"     * @param v the value                                                                  \n"+
"     * @exception BadRecord IF f is not one of MetaTFConstants.BaseOffset, MetaTFConstants.Delta,  \n"+ 
"     *                         MetaTFConstants.Default, MetaTFConstants.SectionOffset      \n"+
"     * @exception IOException IF the write fails                                           \n"+
"     * @see MetaTF.MetaTFConstants                                                         \n"+
"    **/                                                                                    \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, long v) throws BadRecordException, IOException {  \n"+
"        Record.putMetadata(llw, sectionNumber, tag, f, i, v);                              \n"+
"    }                                                                                      \n\n"+

"    /**                                                                                    \n"+
"     * Put a record to change the encoding of a field to a stride encoding     \n"+
"     *                                                                                     \n"+
"     * @param llw the writer                                                               \n"+
"     * @param f the field to be changed                                                    \n"+
"     * @param i the encoding which should be one of MetaTFConstants.Stride, MetaTFConstants.SectionStride  \n"+
"     * @param v1 the first component of the value                                          \n"+
"     * @param v2 the second component of the value                                         \n"+
"     * @exception BadRecord IF f is not one of MetaTFConstants.Stride, MetaTFConstants.SectionStride  \n"+
"     * @exception IOException IF the write fails                                           \n"+
"     */                                                                                    \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, long v1, long v2) throws BadRecordException, IOException { \n"+
"        Record.putMetadata(llw, sectionNumber, tag, f, i, v1, v2);                         \n"+
"    }                                                                                      \n\n"+

"    /**                                                                                    \n"+
"     * Put a record to change the encoding of a field to a string-valued default     \n"+
"     *                                                                                     \n"+
"     * @param llw the writer                                                               \n"+
"     * @param f the field to be changed                                                    \n"+
"     * @param i the encoding which should be MetaTFConstants.Default                 \n"+
"     * @param s the string                                                                 \n"+
"     * @exception BadRecord IF f != MetaTFConstants.Default                                \n"+
"     * @exception IOException IF the write fails                                           \n"+
"     */                                                                                    \n"+
"    public static void putMetadata(LowLevelWriter llw, int f, int i, String v) throws BadRecordException, IOException { \n"+
"        Record.putMetadata(llw, sectionNumber, tag, f, i, v);                              \n"+
"    }                                                                                      \n\n"+

"}  // end of class "+fullname+"                                                 ");

       } catch (NoTagException e) {
            out(
"\n    public void putRecord(LowLevelWriter llw) throws IOException { }  \n"+
"}  // end of class "+fullname+"                                         \n\n");
       }

        /* Finished generating <record name>.java */
        try {
            writer.flush();
            writer.close();
        } catch (IOException e) {
            System.err.println(e.getMessage());
        }
       return null;
    }


    public Object visit(NamedField field, Object data) {
        String name = Util.declassify(field.getName(), (Map)knownFields);
        String type = Util.typify(field.getType());
        out(
"    public static final int "+Util.fieldify(name)+" = "+fieldCount+";              ");
        fieldCount++;
        return "public "+type+" "+name;
    }
        
    public Object visit(TagField field, Object data) {
        try {
            out(
"    public static final int TAG = "+fieldCount+";                                  ");     
            fieldCount++;
            return "public static final int tag = "+field.getTag();
        } catch(NoTagException e) {
            String tagName = Util.declassify("tag", (Map)knownFields); // done for side-effect
            out(
"    public static final int "+Util.fieldify(tagName)+" = "+fieldCount+";           ");     
            fieldCount++;
            return "public final int "+tagName;
        }
    }

    public Object visit(RepeatField field, Object data) {
        return (String)field.getRepeat().accept(this, knownFields)+";                       \n"+
               "    "+(String)field.getRepeatable().accept(this, knownFields)+"[]";
    }
        
    public Object visit(TupleField field, Object data) {
        try {
            String name = field.getName();
            String cname
                 = symtab.getSection(sectionNumber).getRecords().containsKey(name) ?
                       Util.classify(name) :
                       Util.classify(symtab.getSection(0).getName())+"."+Util.classify(name);
            String fname = Util.declassify(name, (Map)knownFields);

            /*
             * Visit subfields just to get the const names
             */
            Iterator it = field.getFields().iterator();
            while(it.hasNext()) {
                ((Field)it.next()).accept(this, knownFields);
            }

            return "public "+cname+" "+fname;
        } catch (DTDException e) {     
            Panic.panic(e.getMessage());
            return null;
        }
    }

    // NOT USED    
    public Object visit(BaseOffset baseOffset, Object data) {
        Panic.panic("ClassGenerator finds BaseOffset");
        return null;
    }    
    public Object visit(Default def, Object data) {
        Panic.panic("ClassGenerator finds Default");
        return null;
    }    
    public Object visit(StringDefault def, Object data) {
        Panic.panic("ClassGenerator finds StringDefault");
        return null;
    }    
    public Object visit(Delta delta, Object data) {
        Panic.panic("ClassGenerator finds Delta");
        return null;
    }    
    public Object visit(None none, Object data) {
        Panic.panic("ClassGenerator finds None");
        return null;
    }    
    public Object visit(NumberValue numberValue, Object data) {
        Panic.panic("ClassGenerator finds NumberValue");
        return null;
    }    
    public Object visit(SectionStride sectionStride, Object data) {
        Panic.panic("ClassGenerator finds SectionStride");
        return null;
    }    
    public Object visit(Stride stride, Object data) {
        Panic.panic("ClassGenerator finds Stride");
        return null;
    }    
    public Object visit(SectionOffset value, Object data) {
        Panic.panic("ClassGenerator finds SectionOffset");
        return null;
    }    

    /**
     * Utility to emit code
     *
     * @param ln one or more lines of code
    **/
    private void out(String ln) {
        try {
            writer.write(ln+"\n"); 
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }


}// end of class ClassGenerator


