/*
    $File$
    Date        Author          Changes
    Nov 21 2000 Richard Jones   Created

    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */

package MetaTF.Generator.Java;

import MetaTF.Parser.*;
import MetaTF.Panic;
import MetaTF.Generator.Java.Util;
import java.util.*;
import java.io.*;

/**********************************************************************************************
 * PutRecordGenerator.
 * Generate the code for writing records
 *     public void putRecord(LowLevelWriter stream) throws BadRecordException, IOException;
 *
 * @see <a href="../../Parser/metatf.jjt">The MetaTF Grammar</a>
 * @version rel-1-2 ($Revision: 1.2 $, $Date: 2000/12/15 16:29:12 $),
 * @author Richard Jones
**/
class PutRecordGenerator implements DTDVisitor {

    /** Field number: used by array types **/
    private int fno;

    /** Prefix such as first part of fully qualified name **/
    String prefix = "";

    /** Suffix such as array index **/
    String suffix = "";

    /**
     * Allocate a new object to write putRecord methods
     *
     * @param w the FileWrite to which the method definition is to be written
    **/
    PutRecordGenerator(FileWriter w) {
        writer = w;
        if(w==null)
            Panic.panic("out: setWriter must be called before out is called");        
    }
    
    /** Reset field numbers, prefixes and suffixes **/
    void resetFieldNumber() {
    	fno = 0;
    	prefix = "";
    	suffix = "";
    }

    /*
     * @param field the named field
     * @param knownFields a Map <name, Integer> of field names and their occurrences
     *                    used to encode distinct names for members
     * @return the full name of this member
    **/
    public Object visit(NamedField field, Object knownFields) {
        String name = Util.declassify(field.getName(), (Map)knownFields);
        String type = Util.kindify(field.getType());
	if(type.equals("Index")) {
	    out(
"        tf = tr.getTraceField("+fno+");                                                        \n"+
"        write(llw, (IndexTraceField)tf, "+prefix+name+suffix+".index);                           ");
        } else {
            out(
"        tf = tr.getTraceField("+fno+");                                                        \n"+
"        write(llw, ("+type+"TraceField)tf, "+prefix+name+suffix+");                           ");
        }

        fno++;
        return prefix+name+suffix;
    }
    
    /*
     * @param field the tag field
     * @param knownFields a Map <name, Integer> of field names and their occurrences
     *                    used to encode distinct names for members
     * @return the full name of this member
    **/
    public Object visit(TagField field, Object knownFields) {
        out(
"        tf = tr.getTraceField("+fno+");                                                        \n"+
"        write(llw, (NumTraceField)tf, "+prefix+"tag"+suffix+");                                              ");

        fno++;
        return prefix+"tag"+suffix;
    }
    
    /*
     * @param field the repeat field
     * @param knownFields a Map <name, Integer> of field names and their occurrences
     *                    used to encode distinct names for members
     * @return null - repeat fields won't be used further
    **/
    public Object visit(RepeatField field, Object knownFields) {
        String repeat = (String)field.getRepeat().accept(this, knownFields);
        if(repeat==null)
            Panic.panic("RepeatField has bad repeat field");
        out(
"        for(int i=0; i<"+repeat+"; i++) {                                             ");

        String oldSuffix = suffix;
        suffix = suffix+"[i]";
        field.getRepeatable().accept(this, knownFields);
        suffix = oldSuffix;
        out(
"        }                                                                             ");

        return null;
    }

    /*
     * @param field the tuple field
     * @param knownFields a Map <name, Integer> of field names and their occurrences
     *                    used to encode distinct names for members
     * @return null - tuple fields won't be used further
    **/    
    public Object visit(TupleField field, Object knownFields) {
    	String name = field.getName();
    	String oldPrefix = prefix;
    	prefix = prefix+Util.declassify(name, (Map) knownFields)+".";
    	Object oldKnownFields = knownFields;	// Tuple field names start from scratch
    	knownFields = new HashMap();
        Iterator it = field.getFields().iterator();
        while(it.hasNext()) {
            Field f = (Field)it.next();
            f.accept(this, knownFields);
        }
        knownFields = oldKnownFields;
        prefix = oldPrefix;
        return null;
    }

        // NOT USED
	public Object visit(DTD symtab, Object data) {
        Panic.panic("PutRecordGenerator finds DTD");
        return null;
    }    
	public Object visit(Section section, Object data) {
        Panic.panic("PutRecordGenerator finds Section");
        return null;
    }    
	public Object visit(Record record, Object data) {
        Panic.panic("PutRecordGenerator finds Record");
        return null;
    }    
	public Object visit(BaseOffset baseOffset, Object data) {
        Panic.panic("PutRecordGenerator finds BaseOffset");
        return null;
    }    
	public Object visit(Default def, Object data) {
        Panic.panic("PutRecordGenerator finds Default");
        return null;
    }    
	public Object visit(StringDefault def, Object data) {
        Panic.panic("PutRecordGenerator finds StringDefault");
        return null;
    }    
	public Object visit(Delta delta, Object data) {
        Panic.panic("PutRecordGenerator finds Delta");
        return null;
    }    
	public Object visit(None none, Object data) {
        Panic.panic("PutRecordGenerator finds None");
        return null;
    }    
	public Object visit(NumberValue numberValue, Object data) {
        Panic.panic("PutRecordGenerator finds NumberValue");
        return null;
    }    
	public Object visit(SectionStride sectionStride, Object data) {
        Panic.panic("PutRecordGenerator finds SectionStride");
        return null;
    }    
	public Object visit(Stride stride, Object data) {
        Panic.panic("PutRecordGenerator finds Stride");
        return null;
    }    
	public Object visit(SectionOffset value, Object data) {
        Panic.panic("PutRecordGenerator finds SectionOffset");
        return null;
    }    

     /** The FileWriter to which output is to be written **/
    private FileWriter writer;

    /**
     * Utility to emit code
     *
     * @param ln one or more lines oc code
    **/
    private void out(String ln) {
        try {
            writer.write(ln+"\n"); 
        } catch (IOException e) {
            System.err.println(e.getMessage());
            System.exit(1);
        }
    }

} // end of PutRecordGenerator class

