/*
    $File$
    Date        Author          Changes
    Nov 30 2000 Richard Jones   Created
    
    Copyright (c) 2000, Richard Jones
    All rights reserved.
 */
 
package MetaTF.Runtime;

import MetaTF.MetaTFConstants;
import MetaTF.BadRecordException;
import java.io.*;
import java.util.*;

/**
 * Superclass for global and user-defined sections                                     
 *                                                                                     
 * Provides:                                                                           
 * 1. putMetadata(current section number, field, encoding [, values...]) methods 
 *    to change encodings of all instances of given field in this section        
 *    before emitting the appropriate metadata record.                                 
 *
 * @version rel-1-2 ($Revision: 1.2 $, $Date: 2001/03/29 15:53:24 $),
 * @author Richard Jones
**/
public abstract class Section {

    /**
     * The lookupField data structure that supports changing all
     * encodings of a given field name in a section.
     * Implemented as List <section><field name>
    **/
    static List[][] lookupField;                                                                

    public static void debug() {
        for(int i=0; i<lookupField.length; i++) {
            System.err.println("Section "+i+":");
            for(int j=0; j<lookupField[i].length; j++) {
                System.err.print(j+": ");
                Iterator it = lookupField[i][j].iterator();
                while(it.hasNext()) {
                    TraceField tf = (TraceField)it.next();
                    System.err.print(tf.name+" ");
                }
                System.err.println();
            }
        }
    }

    /**
     * Initialise the lookupField data structure that support
     * metadata changes to every field of a given name in a section.
     *
     * @param n the number of sections
     * @param f the number of field names
    **/
    static void initLookup(int n, int f) {
        lookupField = new List [n][f];
    }

    /**
     * Start a new list for a name in this section
     *
     * @param s the section number
     * @param n the encoded number of the field name
    **/
    public static void initLookupList(int s, int n) {
        lookupField[s][n] = new ArrayList();
    }

    /**
     * Add a TraceField to the lookupField list
     *
     * @param s the section number
     * @param f the encoded field number
     * @param tf the TraceField
    **/
    static void addLookupField(int s, int f, TraceField tf) {
        lookupField[s][f].add(tf);
    }

    protected static void putSectionMetadata(LowLevelWriter llw, int sectionNumber, int f, int i) throws BadRecordException, IOException {
        changeInterpretation(sectionNumber, f, i);                                                 
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(MetaTFConstants.All);                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.writeln();                                                                 
    }                                                                                  
    protected static void putSectionMetadata(LowLevelWriter llw, int sectionNumber, int f, int i, long v) throws BadRecordException, IOException { 
        changeInterpretation(sectionNumber, f, i, v);                                              
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(MetaTFConstants.All);                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v);                                                                  
        llw.writeln();                                                                 
    }                                                                                  
    protected static void putSectionMetadata(LowLevelWriter llw, int sectionNumber, int f, int i, long v1, long v2) throws BadRecordException, IOException { 
        changeInterpretation(sectionNumber, f, i, v1, v2);                                         
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(MetaTFConstants.All);                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v1);                                                                 
        llw.write(v2);                                                                 
        llw.writeln();                                                                 
    }                                                                                  
    protected static void putSectionMetadata(LowLevelWriter llw, int sectionNumber, int f, int i, String v) throws BadRecordException, IOException { 
        changeInterpretation(sectionNumber, f, i, v);                                              
        llw.write(MetaTFConstants.Metadata);                                           
        llw.write(MetaTFConstants.All);                                                
        llw.write(f);                                                                  
        llw.write(i);                                                                  
        llw.write(v);                                                                  
        llw.writeln();                                                                 
    }

    /** Warning message if changeInterpretation fails **/                        
    private static void warning(String e) { System.err.println("Warning! "+e); }        

    /**                                                                          
    * Change the encodings of all fields of this name.                    
    * @param section the number of the section                                  
    * @param field the encoding of the field to be changed                      
    **/                                                                          
    static void changeInterpretation(int section, int field, int encoding) throws BadRecordException {   
        try {                                                                    
            Iterator it = lookupField[section][field].iterator();                
            while(it.hasNext()) {                                                
                TraceField tf =(TraceField)it.next();                            
                try {                                                            
                    tf.setInterpretationType(encoding);                    
                } catch (BadRecordException e) {                                 
                    warning("Couldn't change encoding: "+e.getMessage()); 
                }                                                                
            }                                                                    
        } catch(ArrayIndexOutOfBoundsException e) {                              
            throw new BadRecordException("Bad value for field in metadata change encoding record");  
        }                                                                        
    }                                                                            
    /**                                                                          
     * Change the encodings of all fields of this name.                    
     * @param section the number of the section                                  
     * @param field the encoding of the field to be changed                      
    **/                                                                          
    static void changeInterpretation(int section, int field, int encoding, long v1) throws BadRecordException {
        try {                                                                    
            Iterator it = lookupField[section][field].iterator();                
            while(it.hasNext()) {                                                
                TraceField tf =(TraceField)it.next();                            
                try {                                                            
                    tf.setInterpretationType(encoding);                    
                    tf.setValue(v1);                                             
                } catch (BadRecordException e) {                                 
                    warning("Couldn't change encoding: "+e.getMessage()); 
                }                                                                
            }                                                                    
        } catch(ArrayIndexOutOfBoundsException e) {                              
            throw new BadRecordException("Bad value for field in metadata change encoding record");  
        }                                                                        
    }                                                                            
    /**                                                                          
     * Change the encodings of all fields of this name.                    
     * @param section the number of the section                                  
     * @param field the encoding of the field to be changed                      
    **/                                                                          
    static void changeInterpretation(int section, int field, int encoding, String v) throws BadRecordException {   
        try {                                                                    
            Iterator it = lookupField[section][field].iterator();                
            while(it.hasNext()) {                                                
                StringTraceField tf =(StringTraceField)it.next();                
                try {                                                            
                    tf.setInterpretationType(encoding);                    
                    tf.setValue(v);                                              
                } catch (BadRecordException e) {                                 
                    warning("Couldn't change encoding: "+e.getMessage());   
                }                                                                
            }                                                                    
        } catch(ArrayIndexOutOfBoundsException e) {                              
            throw new BadRecordException("Bad value for field in metadata change encoding record");  
        }                                                                        
    }                                                                            
    /**                                                                          
     * Change the encodings of all fields of this name.                    
     * @param section the number of the section                                  
     * @param field the encoding of the field to be changed                      
    **/                                                                          
    static void changeInterpretation(int section, int field, int encoding, long v1, long v2) throws BadRecordException {   
        try {                                                                    
            Iterator it = lookupField[section][field].iterator();                
            while(it.hasNext()) {                                                
                TraceField tf =(TraceField)it.next();                            
                try {                                                            
                    tf.setInterpretationType(encoding);                    
                    tf.setValue(v1);                                             
                    tf.setValue2(v2);                                            
                } catch (BadRecordException e) {                                 
                    warning("Couldn't change encoding: "+e.getMessage());   
                }                                                                
            }                                                                    
        } catch(ArrayIndexOutOfBoundsException e) {                              
            throw new BadRecordException("Bad value for field in metadata change encoding record");  
        }                                                                        
    }                                                                            

} // end of class Section                                                              

